@@ -27,14 +27,14 @@ Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
 and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
 the previous versions.
 
-Modifications Copyright (C) 2008-2017 Vinay Sajip. All rights reserved.
+Modifications Copyright (C) 2008-2018 Vinay Sajip. All rights reserved.
 
 A unittest harness (test_gnupg.py) has also been added.
 """
 
-__version__ = "0.4.1"
+__version__ = "0.4.3"
 __author__ = "Vinay Sajip"
-__date__  = "$07-Jul-2017 15:09:20$"
+__date__  = "$13-Jun-2018 12:11:43$"
 
 try:
     from io import StringIO
@@ -217,12 +217,26 @@ class Verify(object):
         "TRUST_ULTIMATE" : TRUST_ULTIMATE,
     }
 
+    # for now, just the most common error codes. This can be expanded as and
+    # when reports come in of other errors.
+    GPG_SYSTEM_ERROR_CODES = {
+        1: 'permission denied',
+        35: 'file exists',
+        81: 'file not found',
+        97: 'not a directory',
+    }
+
+    GPG_ERROR_CODES = {
+        11: 'incorrect passphrase',
+    }
+
     def __init__(self, gpg):
         self.gpg = gpg
         self.valid = False
         self.fingerprint = self.creation_date = self.timestamp = None
         self.signature_id = self.key_id = None
         self.username = None
+        self.key_id = None
         self.key_status = None
         self.status = None
         self.pubkey_fingerprint = None
@@ -295,9 +309,26 @@ class Verify(object):
             if key == "UNEXPECTED":
                 self.status = 'unexpected data'
             else:
-                # N.B. there might be other reasons
+                # N.B. there might be other reasons. For example, if an output
+                # file can't  be created - /dev/null/foo will lead to a
+                # "not a directory" error, but which is not sent as a status
+                # message with the [GNUPG:] prefix. Similarly if you try to
+                # write to "/etc/foo" as a non-root user, a "permission denied"
+                # error will be sent as a non-status message.
+                message = 'error - %s' % value
+                parts = value.split()
+                if parts[-1].isdigit():
+                    code = int(parts[-1])
+                    system_error = bool(code & 0x8000)
+                    code = code & 0x7FFF
+                    if system_error:
+                        mapping = self.GPG_SYSTEM_ERROR_CODES
+                    else:
+                        mapping = self.GPG_ERROR_CODES
+                    if code in mapping:
+                        message = mapping[code]
                 if not self.status:
-                    self.status = 'incorrect passphrase'
+                    self.status = message
         elif key in ("DECRYPTION_INFO", "PLAINTEXT", "PLAINTEXT_LENGTH",
                      "NO_SECKEY", "BEGIN_SIGNING"):
             pass
@@ -409,6 +440,13 @@ class SendResult(object):
     def handle_status(self, key, value):
         logger.debug('SendResult: %s: %s', key, value)
 
+def _set_fields(target, fieldnames, args):
+    for i, var in enumerate(fieldnames):
+        if i < len(args):
+            target[var] = args[i]
+        else:
+            target[var] = 'unavailable'
+
 class SearchKeys(list):
     ''' Handle status messages for --search-keys.
 
@@ -428,11 +466,7 @@ class SearchKeys(list):
 
     def get_fields(self, args):
         result = {}
-        for i, var in enumerate(self.FIELDS):
-            if i < len(args):
-                result[var] = args[i]
-            else:
-                result[var] = 'unavailable'
+        _set_fields(result, self.FIELDS, args)
         result['uids'] = []
         result['sigs'] = []
         return result
@@ -470,7 +504,7 @@ class ListKeys(SearchKeys):
     '''
 
     UID_INDEX = 9
-    FIELDS = 'type trust length algo keyid date expires dummy ownertrust uid sig'.split()
+    FIELDS = 'type trust length algo keyid date expires dummy ownertrust uid sig cap issuer flag token hash curve compliance updated origin'.split()
 
     def __init__(self, gpg):
         super(ListKeys, self).__init__(gpg)
@@ -500,14 +534,25 @@ class ListKeys(SearchKeys):
             self.curkey['subkeys'][-1].append(fp)
             self.key_map[fp] = self.curkey
 
+    def _collect_subkey_info(self, curkey, args):
+        info_map = curkey.setdefault('subkey_info', {})
+        info = {}
+        _set_fields(info, self.FIELDS, args)
+        info_map[args[4]] = info
+
     def sub(self, args):
+        # See issue #81. We create a dict with more information about
+        # subkeys, but for backward compatibility reason, have to add it in
+        # as a separate entry 'subkey_info'
         subkey = [args[4], args[11]]    # keyid, type
         self.curkey['subkeys'].append(subkey)
+        self._collect_subkey_info(self.curkey, args)
         self.in_subkey = True
 
     def ssb(self, args):
         subkey = [args[4], None]    # keyid, type
         self.curkey['subkeys'].append(subkey)
+        self._collect_subkey_info(self.curkey, args)
         self.in_subkey = True
 
     def sig(self, args):
@@ -522,6 +567,7 @@ class ScanKeys(ListKeys):
         # use the last value args[-1] instead of args[11]
         subkey = [args[4], args[-1]]
         self.curkey['subkeys'].append(subkey)
+        self._collect_subkey_info(self.curkey, args)
         self.in_subkey = True
 
 class TextHandler(object):
@@ -544,6 +590,7 @@ class Crypt(Verify, TextHandler):
         self.data = ''
         self.ok = False
         self.status = ''
+        self.key_id = None
 
     def __nonzero__(self):
         if self.ok: return True
@@ -580,7 +627,10 @@ class Crypt(Verify, TextHandler):
             self.status = 'sig created'
         elif key == "SIGEXPIRED":  # pragma: no cover
             self.status = 'sig expired'
-        elif key in ("ENC_TO", "USERID_HINT", "GOODMDC",
+        elif key == "ENC_TO":  # pragma: no cover
+            # ENC_TO <long_keyid> <keytype> <keylength>
+            self.key_id = value.split(' ', 1)[0]
+        elif key in ("USERID_HINT", "GOODMDC",
                      "END_DECRYPTION", "CARDCTRL", "BADMDC",
                      "SC_OP_FAILURE", "SC_OP_SUCCESS",
                      "PINENTRY_LAUNCHED", "KEY_CONSIDERED"):
@@ -662,6 +712,8 @@ class Sign(TextHandler):
         self.hash_algo = None
         self.fingerprint = None
         self.status = None
+        self.key_id = None
+        self.username = None
 
     def __nonzero__(self):
         return self.fingerprint is not None
@@ -677,17 +729,19 @@ class Sign(TextHandler):
             self.status = 'key revoked'
         elif key == "SIG_CREATED":
             (self.type,
-             algo, self.hash_algo, cls,
-             self.timestamp, self.fingerprint
-             ) = value.split()
+             algo, self.hash_algo, cls, self.timestamp, self.fingerprint
+            ) = value.split()
             self.status = 'signature created'
-        elif key in ("USERID_HINT", "NEED_PASSPHRASE", "GOOD_PASSPHRASE",
-                     "BAD_PASSPHRASE", "BEGIN_SIGNING"):
+        elif key == "USERID_HINT":  # pragma: no cover
+            self.key_id, self.username = value.split(' ', 1)
+        elif key == "BAD_PASSPHRASE":
+            self.status = 'bad passphrase'
+        elif key in ("NEED_PASSPHRASE", "GOOD_PASSPHRASE", "BEGIN_SIGNING"):
             pass
         else:  # pragma: no cover
             logger.debug('message ignored: %s, %s', key, value)
 
-VERSION_RE = re.compile(r'gpg \(GnuPG\) (\d+(\.\d+)*)'.encode('ascii'), re.I)
+VERSION_RE = re.compile(r'gpg \(GnuPG(?:/MacGPG2)?\) (\d+(\.\d+)*)'.encode('ascii'), re.I)
 HEX_DIGITS_RE = re.compile(r'[0-9a-f]+$', re.I)
 
 class GPG(object):
@@ -754,7 +808,7 @@ class GPG(object):
         try:
             p = self._open_subprocess(["--version"])
         except OSError:
-            msg = 'Unable to run gpg - it may not be available.'
+            msg = 'Unable to run gpg (%s) - it may not be available.' % self.gpgbinary
             logger.exception(msg)
             raise OSError(msg)
         result = self.result_map['verify'](self) # any result will do for this
@@ -775,8 +829,9 @@ class GPG(object):
         will be appended. The ``passphrase`` argument needs to be True if
         a passphrase will be sent to GPG, else False.
         """
-        cmd = [self.gpgbinary, '--status-fd', '2', '--no-tty']
-        cmd.extend(['--debug', 'ipc'])
+        cmd = [self.gpgbinary, '--status-fd', '2', '--no-tty', '--no-verbose']
+        if 'DEBUG_IPC' in os.environ:
+            cmd.extend(['--debug', 'ipc'])
         if passphrase and hasattr(self, 'version'):
             if self.version >= (2, 1):
                 cmd[1:1] = ['--pinentry-mode', 'loopback']
@@ -1094,10 +1149,20 @@ class GPG(object):
         data.close()
         return result
 
-    def delete_keys(self, fingerprints, secret=False, passphrase=None):
+    def delete_keys(self, fingerprints, secret=False, passphrase=None,
+                    expect_passphrase=True):
+        """
+        Delete the indicated keys.
+
+        Since GnuPG 2.1, you can't delete secret keys without providing a
+        passphrase. However, if you're expecting the passphrase to go to gpg
+        via pinentry, you should specify expect_passphrase=False. (It's only
+        checked for GnuPG >= 2.1).
+        """
         which='key'
         if secret:  # pragma: no cover
-            if self.version >= (2, 1) and passphrase is None:
+            if (self.version >= (2, 1) and passphrase is None and
+                expect_passphrase):
                 raise ValueError('For GnuPG >= 2.1, deleting secret keys '
                                  'needs a passphrase to be provided')
             which='secret-key'
@@ -1122,18 +1187,21 @@ class GPG(object):
         return result
 
     def export_keys(self, keyids, secret=False, armor=True, minimal=False,
-                    passphrase=None):
+                    passphrase=None, expect_passphrase=True):
         """
         Export the indicated keys. A 'keyid' is anything gpg accepts.
 
         Since GnuPG 2.1, you can't export secret keys without providing a
-        passphrase.
+        passphrase. However, if you're expecting the passphrase to go to gpg
+        via pinentry, you should specify expect_passphrase=False. (It's only
+        checked for GnuPG >= 2.1).
         """
 
         which=''
         if secret:
             which='-secret-key'
-            if self.version >= (2, 1) and passphrase is None:
+            if (self.version >= (2, 1) and passphrase is None and
+                expect_passphrase):
                 raise ValueError('For GnuPG >= 2.1, exporting secret keys '
                                  'needs a passphrase to be provided')
         if _is_sequence(keyids):
@@ -1458,3 +1526,30 @@ class GPG(object):
         self._handle_io(args, file, result, passphrase, binary=True)
         logger.debug('decrypt result: %r', result.data)
         return result
+
+    def trust_keys(self, fingerprints, trustlevel):
+        levels = Verify.TRUST_LEVELS
+        if trustlevel not in levels:
+            poss = ', '.join(sorted(levels))
+            raise ValueError('Invalid trust level: "%s" (must be one of %s)' %
+                             (trustlevel, poss))
+        trustlevel = levels[trustlevel] + 2
+        import tempfile
+        try:
+            fd, fn = tempfile.mkstemp()
+            lines = []
+            if isinstance(fingerprints, string_types):
+                fingerprints = [fingerprints]
+            for f in fingerprints:
+                lines.append('%s:%s:' % (f, trustlevel))
+            # The trailing newline is required!
+            s = os.linesep.join(lines) + os.linesep
+            logger.debug('writing ownertrust info: %s', s);
+            os.write(fd, s.encode(self.encoding))
+            os.close(fd)
+            result = self.result_map['delete'](self)
+            p = self._open_subprocess(['--import-ownertrust', fn])
+            self._collect_output(p, result, stdin=p.stdin)
+        finally:
+            os.remove(fn)
+        return result
